from pydantic import BaseModel, Field
import uuid
import blinker
from dataclasses import dataclass, field
from typing import TypeVar, Generic, Callable

T = TypeVar("T")

@dataclass
class SignalResponse:
    request_id: str = field(default_factory=lambda: str(uuid.uuid4()))

class CoolSignal(blinker.Signal, Generic[T]):

    @staticmethod
    def _wrap_handler(func: Callable) -> Callable:
        def wrapper(sender, *args, **kwargs):
            request_id = None
            for arg in args:
                if isinstance(arg, SignalResponse):
                    request_id = arg.request_id
                    break
            if request_id is None:
                for arg in kwargs.values():
                    if isinstance(arg, SignalResponse):
                        request_id = arg.request_id
                        break
            arg = func(sender, *args, **kwargs)
            return (request_id, arg)
        wrapper._is_wrapped_for_request_id = True
        return wrapper
    
    def connect(self, receiver: Callable, sender=blinker.ANY, weak=False):
        if not getattr(receiver, "_is_wrapped_for_request_id", False):
            receiver = self._wrap_handler(receiver)
        return super().connect(receiver, sender=sender, weak=weak)
    
    def send(self, sender=None, **kwargs):
        responses = super().send(sender, **kwargs)
        request_id = None
        for value in kwargs.values():
            if isinstance(value, SignalResponse):
                request_id = value.request_id
                break
        if request_id is not None:
            for _, (rid, value) in responses:
                if rid == request_id and value is not None:
                    return value
            return None
        return responses



from src.core import CoolSignal, SignalResponse


def test_send_returns_all_responses_when_no_request_id():
    # arrange
    signal = CoolSignal()

    def handler(sender, **kwargs):
        return "ok"

    # act
    signal.connect(handler)
    response = signal.send(signal)

    # assert
    assert isinstance(response, list)
    assert response[0][1] == (None, "ok")


def test_send_returns_none_for_non_matching_request_id():
    # arrange
    signal = CoolSignal()
    response = SignalResponse(request_id="not-matching")

    def handler(sender, response: SignalResponse):
        if response.request_id == "some-other-id":
            return "should not match"

    # act
    signal.connect(handler)
    response = signal.send(signal, response=response)

    # assert
    assert response is None


def test_send_returns_value_for_matching_request_id_among_multiple_handlers():
    # arrange
    signal = CoolSignal()
    response1 = SignalResponse()
    response2 = SignalResponse()
    response3 = SignalResponse()

    def handler1(sender, response: SignalResponse):
        if response.request_id == response1.request_id:
            return "handler1"

    def handler2(sender, response: SignalResponse):
        if response.request_id == response2.request_id:
            return "handler2"

    def handler3(sender, response: SignalResponse):
        if response.request_id == response3.request_id:
            return "handler3"

    # act
    signal.connect(handler1)
    signal.connect(handler2)
    signal.connect(handler3)

    # assert
    assert signal.send(signal, response=response1) == "handler1"
    assert signal.send(signal, response=response2) == "handler2"
    assert signal.send(signal, response=response3) == "handler3"
